#!/usr/bin/perl -w

use strict;
use Fcntl ':flock';
use Cwd qw(abs_path);
use FindBin;

# Asks for confirmation from user.
sub confirm($) {
	print $_[0];
	my $reply = <STDIN>;
	if (($reply !~ m/^y$/i) and ($reply !~ m/^yes$/i)) {
		exit(2);
	}
}

# Reads the file contents and returns it (silently returns '' in case of error).
sub read_file($) {
	my ($filename) = @_;
	my $text = '';
	my $fh;

	if (open($fh, $filename)) {
		local $/ = undef;
		$text = <$fh>;
		close($fh);
	}
	return $text;
}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Read command-line parameters (port and optional sudo command).
my $port = shift;
my $port_specified;
if ($port) {
	if (($port !~ m/^\d+$/) or ($port > 65535)) {
		die "    Error: Invalid port '$port' specified!\n";
	}
	$port_specified = 1;
}
else {
	$port_specified = 0;
}

my $sudo_cmd = shift;


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Build list of servers.
my %SERVER_TEMPLATE = (
	'PORT'     => 0,            # port used by the server
	'PID'      => 0,            # PID of the server process
	'UID'      => 0,            # UID of the server process
	'USERNAME' => '',           # username for the UID
	'BRANCH'   => ''            # is it Linux or Moblin AppChecker
);
my %servers_by_port = ();       # port => SERVER
my %servers_by_pid = ();        # PID => SERVER
my %servers_by_uid = ();        # UID => (SERVER*)

# Read the list of lock files.
opendir(DIR, '/tmp') or die "    Error: Could not access /tmp!\n";
my @lock_list = sort(grep(m/^app-checker-server\.lock\.\d+$/, readdir(DIR)));
closedir(DIR);
foreach my $lock_file (@lock_list) {
	if (open(LOCKFILE, "/tmp/$lock_file")) {
		# Process the file only if it is locked (that is, server is really running).
		if (!flock(LOCKFILE, LOCK_EX | LOCK_NB)) {
			my %SERVER = %SERVER_TEMPLATE;
			my $pid = <LOCKFILE>;
			$pid =~ s/^[LG]:(\d+)$/$1/;
			$SERVER{'PID'} = $pid;
			$lock_file =~ m/\.lock\.(\d+)/;
			$SERVER{'PORT'} = $1;
			$SERVER{'UID'} = (stat(LOCKFILE))[4];   # Use the lock file's owner to know the process UID.
			$SERVER{'USERNAME'} = getpwuid($SERVER{'UID'});
			my $cmdline = read_file("/proc/$pid/cmdline");
			if ($cmdline =~ m!/opt/moblin/app-checker/|/moblin-app-checker/!) {
				$SERVER{'BRANCH'} = 'Moblin AppChecker';
			}
			elsif ($cmdline =~ m!/opt/lsb/app-checker/|/linux-app-checker/!) {
				$SERVER{'BRANCH'} = 'Linux AppChecker';
			}
			else {
				$SERVER{'BRANCH'} = '<unknown>';
			}
			$servers_by_port{$SERVER{'PORT'}} = \%SERVER;
			$servers_by_pid{$SERVER{'PID'}} = \%SERVER;
			if (!$servers_by_uid{$SERVER{'UID'}}) {
				$servers_by_uid{$SERVER{'UID'}} = [];
			}
			push @{$servers_by_uid{$SERVER{'UID'}}}, \%SERVER;
		}
		else {
			flock(LOCKFILE, LOCK_UN);
		}
		close(LOCKFILE);
	}
}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Show prompt.
if ($port_specified) {
	# The port was specified explicitly.
	if ($servers_by_port{$port}) {
		# There is a server running on this port.
		if (($servers_by_port{$port}->{'UID'} == $>) or ($> == 0) or $sudo_cmd) {
			# 1. This server is owned by the current user, or
			# 2. We are root, or
			# 3. sudo command will be used.
			# In any of these cases we have permissions to stop the server, so continue.
			confirm('    Are you sure you want to stop the server? ');
		}
		else {
			# Most probably, we will be unable to kill the process, but who knows...
			print "    The server is owned by the user '".$servers_by_port{$port}->{'USERNAME'}."' and you may not have permissions to stop it.";
			$sudo_cmd = 'su -c';
		}
	}
	else {
		die "    No AppChecker server using the port '$port' was found.\n";
	}
}
else {
	# Port was not specified.
	my @list_servers;
	if (($> == 0) or $sudo_cmd) {
		# We have persmissions to stop all servers, list them all.
		@list_servers = sort { $a->{'PORT'} <=> $b->{'PORT'} } values(%servers_by_port);
		if (scalar(@list_servers) == 0) {
			# No servers found.
			die "    No running AppChecker servers found.\n";
		}
	}
	else {
		# List only servers started by the current user.
		@list_servers = sort { $a->{'PORT'} <=> $b->{'PORT'} } @{$servers_by_uid{$>}};
		if (scalar(@list_servers) == 0) {
			# No servers found.
			die "    There are no running AppChecker servers started by you.\n";
		}
	}

	if (scalar(@list_servers) == 1) {
		# Only one server was found.
		confirm('    Are you sure you want to stop the server? ');
		$port = $list_servers[0]->{'PORT'};
	}
	else {
		# List the servers and ask which to stop.
		my $i = 0;
		while ($i < scalar(@list_servers)) {
			my $username = $list_servers[$i]->{'USERNAME'};
			if ($username eq 'moblinappchecker') {
				$username .= ' (system)';
			}
			print '    ['.($i + 1).']: '.$list_servers[$i]->{'BRANCH'}.'; port '.$list_servers[$i]->{'PORT'}.'; PID: '.$list_servers[$i]->{'PID'}.'; owner: '.$username."\n";
			++$i;
		}
		print '    Please specify the number of the server you want to stop: ';
		$i = <STDIN>;
		$i =~ s/[\r\n]//g;
		--$i;
		if (($i < 0) or ($i >= scalar(@list_servers))) {
			die "    Wrong number specified.\n";
		}
		$port = $list_servers[$i]->{'PORT'};
	}
}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Stop the server.
my $cmd = abs_path($FindBin::Bin.'/../utils/stop_tests.pl');
$cmd =~ s/([ ()\'])/\\$1/g;
$cmd = $cmd.' '.$servers_by_port{$port}->{'PID'};

if ($sudo_cmd) {
	print "\n    The command '$sudo_cmd' will be used to gain root access.\n    If you want to change this, run $0 <port> <sudo-command>\n\n";
	$cmd = "$sudo_cmd '$cmd'";
}

system($cmd);
exit($?);
