#!/usr/bin/perl -w

# Moblin Application Checker
# General Templates and Functions Module (Templates.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use strict;
use File::Temp qw/tempdir/;

my $version = '0.2';
my $listfile;                   # List file containing all the packages to be tested
my $cmdline = "$0 ".join(' ', @ARGV);

# Turn autoflush on
my $oldh = select(STDOUT);
$| = 1;
select(STDERR);
$| = 1;
select($oldh);

# Reads next command line parameter, parses it and removes from the @ARGV.
# Parameters:
#   1. (IN)  Ref to array of hashes containing info about options: {LONG_NAME => 'param', SHORT_NAME => 'p', HAS_VALUE => [1|0]}
# Return value:
#   Reference to hash containing info about the next option: {NAME => 'option-name', VALUE => 'value'}
#   If present, NAME contains the long option name, else the short one. If value is not necessary, VALUE contains undef.
#   If parameter is invalid, return value is {NAME => undef, VALUE => undef}.
#   If no more options present, return value is undef.
sub get_option($) {
	my ($options) = @_;
	my $param = shift @ARGV;
	if (!$param) {
		return undef;
	}
	foreach (@$options) {
		my $long_name = $_->{'LONG_NAME'};
		my $short_name = $_->{'SHORT_NAME'};
		my $return_name = ($long_name or $short_name);
		if ($_->{'HAS_VALUE'}) {
			if ($long_name and ($param =~ m/^--$long_name=(.+)$/)) {
				return {'NAME' => $return_name, 'VALUE' => $1};
			}
			elsif ($short_name) {
				if ($param =~ m/^-$short_name(.+)$/) {
				return {'NAME' => $return_name, 'VALUE' => $1};
				}
				elsif ($param =~ m/^-$short_name$/) {
					my $value = shift @ARGV;
					if (!$value) {
						print "    Error: Option $param must be followed by a value!\n";
						return {'NAME' => undef, 'VALUE' => undef};
					}
					else {
						return {'NAME' => $return_name, 'VALUE' => $value};
					}
				}
			}
		}
		else {
			if ($long_name and ($param =~ m/^--$long_name$/)) {
				return {'NAME' => $return_name, 'VALUE' => undef};
			}
			elsif ($short_name and ($param =~ m/^-$short_name$/)) {
				return {'NAME' => $return_name, 'VALUE' => undef};
			}
		}
	}
	print "    Error: Unknown option: $param\n";
	return {'NAME' => undef, 'VALUE' => undef};
}

# Escape variable values for running the shell command.
# Values to be escaped should be enclosed in double angle quotation marks: «...»
# The file is to be opened in UTF-8 encoding, otherwise quotation marks may look different
# (which does not cause any problems if you copy-paste them exactly as they are, since
# processing is octet-based anyway).
# NO special Perl pragmas are needed!
sub esc_cmd($) {
	my ($str) = @_;
	sub e($) {
		my ($s) = @_;
		$s =~ s/\'/\'\\\'\'/g;
		return "'$s'";
	}
	$str =~ s/«(.*?)»/e($1)/eg;
	return $str;
}

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# TET journal functionality
# TODO: If necessary, split it later into separate module or class

my %RESULT_CODES = (
	'PASS' => 0,
	'FAIL' => 1,
	'UNRESOLVED' => 2,
	'NOTINUSE' => 3,
	'UNSUPPORTED' => 4,
	'UNTESTED' => 5,
	'UNINITIATED' => 6,
	'UNREPORTED' => 7,
	'WARNING' => 101,
	'FIP' => 102,
	'NOTIMP' => 103,
	'UNAPPROVE' => 104
);

my $tet_system = `uname -snrvm`;
$tet_system =~ s/[\r\n]//g;
my $tet_activity = 0;
my $tet_tp_number = 0;

sub format_date_time() {
	my ($sec, $min, $hour, $mday, $mon, $year) = localtime();
	return sprintf('%02d:%02d:%02d %04d%02d%02d', $hour, $min, $sec, $year + 1900, $mon + 1, $mday);
}

sub format_time() {
	my ($sec, $min, $hour, $mday, $mon, $year) = localtime();
	return sprintf('%02d:%02d:%02d', $hour, $min, $sec);
}

sub tet_header($) {
	my ($HJOURNAL) = @_;
	print $HJOURNAL "0|3.7-lite ".format_date_time()."|User: ".getpwuid($<)." ($<) TCC Start, Command line: $cmdline\n";
	print $HJOURNAL "5|$tet_system|System Information\n";
	print $HJOURNAL "30||VSX_NAME=moblinpkgchk.pl $version\n";
	print $HJOURNAL "40||Config End\n";
}

sub tet_tc_start($$) {
	my ($HJOURNAL, $file) = @_;
	$tet_activity = (defined($tet_activity) ? ($tet_activity + 1) : 0);
	print $HJOURNAL "10|$tet_activity $file ".format_time()."|TC Start\n";
	print $HJOURNAL "15|$tet_activity tetj-1.0 1|TCM Start\n";
}

sub tet_tp_start($;$) {
	my ($HJOURNAL, $tp_info) = @_;
	$tet_tp_number = (defined($tet_tp_number) ? ($tet_tp_number + 1) : 0);
	print $HJOURNAL "400|$tet_activity $tet_tp_number 1 ".format_time()."|IC Start\n";
	print $HJOURNAL "200|$tet_activity $tet_tp_number ".format_time()."|".($tp_info or 'TP Start')."\n";
}

sub tet_tp_result($$;$) {
	my ($HJOURNAL, $result, $info) = @_;
	if (defined($info)) {
		foreach my $line (split(m/\r?\n/, $info)) {
			print $HJOURNAL "520|$tet_activity $tet_tp_number 0 0 0|$line\n";
		}
	}
	print $HJOURNAL "220|$tet_activity $tet_tp_number ".$RESULT_CODES{$result}." ".format_time()."|$result\n";
}

sub tet_tp_end($) {
	my ($HJOURNAL) = @_;
	print $HJOURNAL "410|$tet_activity $tet_tp_number 1 ".format_time()."|IC End\n";
}

sub tet_tc_end($) {
	my ($HJOURNAL) = @_;
	print $HJOURNAL "80|$tet_activity 0 ".format_time()."|TC End\n";
}

sub tet_footer($) {
	my ($HJOURNAL) = @_;
	print $HJOURNAL "900|".format_time()."|TCC End\n";
}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Main functionality

# Prints the help text
sub help() {
	print <<DATA;
    Usage: $0 [OPTION]...
    -l <list-file>, --list=<list-file>
                            Specifies the file containing the list of packages
                            to check. Format:
                            <package> TAB <journal> [ TAB <unpacked-dir> ]
    -v, --version           Prints the version number of the program.
    -h, --help              Display this help and exit.
DATA
}

my @cmd_line_options = (
	{'LONG_NAME' => 'help',    'SHORT_NAME' => 'h', 'HAS_VALUE' => 0},
	{'LONG_NAME' => 'version', 'SHORT_NAME' => 'v', 'HAS_VALUE' => 0},
	{'LONG_NAME' => 'list',    'SHORT_NAME' => 'l', 'HAS_VALUE' => 1}
);

# Read and process the command-line parameters
while (my $res = get_option(\@cmd_line_options)) {
	if (!$res->{'NAME'}) {
		next;
	}
	if ($res->{'NAME'} eq 'help') {
		help();
		exit;
	}
	elsif ($res->{'NAME'} eq 'version') {
		print <<DATA;
Moblin Package Check version $version
Copyright (C) 2009 The Linux Foundation. All rights reserved.
DATA
		exit;
	}
	elsif ($res->{'NAME'} eq 'list') {
		$listfile = $res->{'VALUE'};
	}
}

if (!$listfile) {
	help();
	exit;
}

# List of directories to be cleaned before the program exits
my @cleanup_dirs = ();

# Function for unpacking packages into temporary directory
sub unpack_package($) {
	my ($pkg) = @_;
	my $unpacked_dir = tempdir('moblin-pkgchk-XXXXXX', TMPDIR => 1);
	push @cleanup_dirs, $unpacked_dir;
	if ($pkg =~ m/\.rpm$/) {
		system(esc_cmd("cd «$unpacked_dir» && rpm2cpio «$pkg» | cpio -id"));
		if ($?) {
			$unpacked_dir = '';
		}
	}
	elsif ($pkg =~ m/\.deb$/) {
		system(esc_cmd("cd «$unpacked_dir» && ar -x «$pkg»"));
		if ($?) {
			$unpacked_dir = '';
		}
		else {
			if ((-f "$unpacked_dir/data.tar.gz") or (-f "$unpacked_dir/data.tar.bz2")) {
				my $unpacked_dir2 = tempdir('moblin-pkgchk-XXXXXX', TMPDIR => 1);
				push @cleanup_dirs, $unpacked_dir2;
				if (-f "$unpacked_dir/data.tar.gz") {
					system(esc_cmd("cd «$unpacked_dir2» && tar xzf «$unpacked_dir/data.tar.gz»"));
				}
				else {
					system(esc_cmd("cd «$unpacked_dir2» && tar xjf «$unpacked_dir/data.tar.bz2»"));
				}
				if (!$?) {
					$unpacked_dir = $unpacked_dir2;
				}
				else {
					$unpacked_dir = '';
				}
			}
			else {
				$unpacked_dir = '';
			}
		}
	}
	else {
		$unpacked_dir = '';
	}
	return $unpacked_dir;
}

# List of packages to check.
# Format of the elements:
#   'PKG_FILE'     - path to the original package
#   'JOURNAL'      - path to the journal file to be created
#   'UNPACKED_DIR' - path to the directory containing the unpacked package contents
my @packages = ();

# Read and parse the list of packages
open(LISTFILE, $listfile) or die "Failed to open '$listfile' for reading:\n$!";
while (my $line = <LISTFILE>) {
	$line =~ s/[\r\n]//g;
	next if ($line =~ m/^#/);
	if ($line =~ m/^([^\t]+)\t([^\t]+)(?:\t([^\t]+))?/) {
		my ($pkg_file, $journal, $unpacked_dir) = ($1, $2, $3);
		if (!$unpacked_dir) {
			$unpacked_dir = unpack_package($1);
		}
		push @packages, {'PKG_FILE' => $pkg_file, 'JOURNAL' => $journal, 'UNPACKED_DIR' => $unpacked_dir};
	}
}
close(LISTFILE);

# Function for recursive getting each file from the directory and all subdirectories
# and processing it with the callback function
sub recursive_fs_walk($$);
sub recursive_fs_walk($$) {
	my ($path, $callback) = @_;
	if (opendir(DIR, $path)) {
		my @dir_list = grep(!/^\.\.?$/, readdir(DIR));
		closedir(DIR);
		foreach my $item (@dir_list) {
			my $item_full = "$path/$item";
			if (-d $item_full) {
				recursive_fs_walk($item_full, $callback);
			}
			elsif (-f $item_full) {
				$callback->($item_full);
			}
			else {
				print STDERR "Unknown type of '$item_full'! Skipping.";
			}
		}
	}
	else {
		print STDERR "Failed to open directory '$path' for reading:\n$!";
	}
}

# Start checking
foreach my $pkg (@packages) {
	print "TESTING: $pkg->{'PKG_FILE'}\n";
	my $HJOURNAL;
	if (!open($HJOURNAL, ">$pkg->{'JOURNAL'}")) {
		print STDERR "Failed to open journal file '$pkg->{'JOURNAL'}' for writing:\n$!\n";
		next;
	}
	tet_header($HJOURNAL);
	tet_tc_start($HJOURNAL, $pkg->{'PKG_FILE'});

	my $desktop_file_found = 0;
	my $install_in_opt;
	my $root = $pkg->{'UNPACKED_DIR'};
	if ($root) {
		recursive_fs_walk(
			$root,
			sub ($) {
				my ($file) = @_;
				return if (!($file =~ s/^$root//));
				if ($file =~ m/\.desktop$/) {
					$desktop_file_found = 1;
					tet_tp_start($HJOURNAL, "Checking location for $file");
					if ($file =~ m!^/usr/share/applications/!) {
						tet_tp_result($HJOURNAL, 'PASS');
					}
					else {
						tet_tp_result($HJOURNAL, 'FAIL', "Location of the file '$file' is not Moblin compliant! The .desktop file should be installed into /usr/share/applications");
					}
					tet_tp_end($HJOURNAL);
				}
				elsif ($file =~ m/\.so(\.[^\/]*)?$/) { # libraries
					tet_tp_start($HJOURNAL, "Checking location for $file");
					if ($file =~ m!^/opt/!) {
						tet_tp_result($HJOURNAL, 'PASS');
						if (!defined($install_in_opt)) {
							$install_in_opt = $file;
							$install_in_opt =~ s!^/opt/([^/]+)/.*$!$1!;
						}
					}
					else {
						tet_tp_result($HJOURNAL, 'FAIL', "Location of the file '$file' is not Moblin compliant! Libraries provided by the application should be installed into /opt/<domain>/<app name>.");
					}
					tet_tp_end($HJOURNAL);
				}
				else {
					# No requirements for other file types in Moblin spec.
				}
			}
		);
	}
	else {
		tet_tp_start($HJOURNAL, "Unpacking package $pkg->{'PKG_FILE'}");
		tet_tp_result($HJOURNAL, 'FAIL', "Could not unpack the package!");
		tet_tp_end($HJOURNAL);
	}
	if (defined($install_in_opt)) {
		tet_tp_start($HJOURNAL, "Checking LANANA name");
		tet_tp_result($HJOURNAL, 'FIP', "The package should install into /opt/<package> or /opt/<provider> directory. Please, check that the subdirectory name you are using ('$install_in_opt') is either name of your package or a LANANA name registered to you.");
		tet_tp_end($HJOURNAL);
	}
	if (!$desktop_file_found) {
		tet_tp_start($HJOURNAL, "Checking .desktop file existence");
		tet_tp_result($HJOURNAL, 'FAIL', "The package does not contain any .desktop file.");
		tet_tp_end($HJOURNAL);
	}

	tet_tc_end($HJOURNAL);
	tet_footer($HJOURNAL);
	close($HJOURNAL);
}

foreach (@cleanup_dirs) {
	system(esc_cmd("rm -rf «$_» >/dev/null 2>&1"));
}
