package Distros;
use strict;

require Exporter;
@Distros::ISA = qw(Exporter);
# Exported symbols
push @Distros::EXPORT, qw(
    &guess_platform &guess_package_manager &does_shell_know &trim &is_elf_arch_correct
    &calc_abs_path &escape_shell_symbols &check_main_entry
    %known_commands %known_wrong_commands
    %NotUploadableSymbols %AliasedCommands
);

#----------------------------------------------------------------------------

# Helps to determine platform name.
sub guess_platform($ )
{
    my $name = $_[0];

    if ( $name =~ /^(amd64|x86_64)$/)
        { return "x86-64" }
    elsif ( $name =~ /^(ia32|i[3-6]86|x86)$/ )
        { return "IA32" }
    elsif ( $name =~ /^(ia64)$/ )
        { return "IA64" }
    elsif ( $name =~ /^(ppc|ppc32)$/ )
        { return "PPC32" }
    elsif ( $name =~ /^(ppc64)$/ )
        { return "PPC64" }
    elsif ( $name =~ /^(s390)$/ )
        { return "S390" }
    elsif ( $name =~ /^(s390x)$/ )
        { return "S390X" }
    else
        { return "unknown"; }
}

#----------------------------------------------------------------------------

# Determines which package manager to use.
sub guess_package_manager {
    my $host_os = "";
    if( does_shell_know ("lsb_release") ) {
        $host_os = `lsb_release -s -d`;
    }

    $host_os =~ s/\"(.*)\"/$1/;
    chomp($host_os);

    if ( $host_os =~ /(suse|red\W?hat|asianux|mandriva)/i ) {
        # Distros known to be rpm-based
        return "rpm";
    }
    elsif ( $host_os =~ /(ubuntu|debian)/i ) {
        # Distros known to use apt and dpkg
        # (note that on some Debian-based systems, like Xandros, `lsb_release` gives us 'Debian')
        return "dpkg";
    }
    elsif ( does_shell_know ("emerge") ) {
        # Gentoo
        return "emerge";
    }
    elsif ( does_shell_know ("pacman") ) {
        # Arch Linux
        return "pacman";
    }
    else {
        # Do we have 'dpkg'? Some systems may have both rpm and dpkg,
        # and 'dpkg' is usually the default manager in this case
        if ( does_shell_know ("dpkg") ) {
            return "dpkg";
        }
        elsif ( does_shell_know ("rpm") ) {
            return "rpm";
        }
    }

    return "unknown";
}

#----------------------------------------------------------------------------

sub is_elf_arch_correct($$$ ) {
    my($arch, $elf_arch, $elf_arch_bit) = ($_[0],$_[1],$_[2]);

    if( $arch =~ /x86-64/i ) {
        if( $elf_arch !~ /86-64/ ) {
            return 0;
        }
    }
    elsif( $arch =~ /IA64/i ) {
        if( $elf_arch !~ /IA-64/ ) {
            return 0;
        }
    }
    elsif( $arch =~ /S390X/i ) {
        if( ($elf_arch !~ /390/) or ($elf_arch_bit ne "64") ) {
            return 0;
        }
    }
    elsif( $arch =~ /PPC64/i ) {
        if( $elf_arch !~ /PowerPC 64-bit/ and $elf_arch !~ /PowerPC64/ and ($elf_arch !~ /PowerPC/ or $elf_arch_bit ne "64") ) {
            return 0;
        }
    }
    elsif( $arch =~ /PPC32/i ) {
        if( ($elf_arch !~ /PowerPC/) or ($elf_arch =~ /PowerPC 64-bit/) or ($elf_arch =~ /PowerPC64/) ) {
            return 0;
        }
    }
    elsif( $arch =~ /IA32/i ) {
        if( $elf_arch !~ /Intel 80386/ ) {
            return 0;
        }
    }
    elsif( $arch =~ /S390/i ) {
        if( ($elf_arch !~ /390/) or ($elf_arch_bit ne "32") ) {
            return 0;
        }
    }

    return 1;
}

#----------------------------------------------------------------------------

# Checks whether the specified tool is known to the shell.
sub does_shell_know
{
    my $tool_name = shift;
    return (system ("which $tool_name >/dev/null 2>/dev/null") == 0);
}

# Trim string spaces.
sub trim($)
{
    my $string = shift;
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    return $string;
}

# Escape characters that are special to the shell
sub escape_shell_symbols($ )
{
    my $string = shift;
    $string =~ s/([;<>\*\|` &\$!#\(\)\[\]\{\}:'"])/\\$1/g;
    return $string;
}

# Calculate absolute path for symlink target_base
sub calc_abs_path($$ )
{
    my($src, $dst) = ($_[0],$_[1]);
    my $real_dst='';

    if( $dst !~ /^\.\./ ) { # symlink to smth in the current directory
        $dst =~ s/^\.\/?//;
        $real_dst = $src;
        if( $real_dst =~ /(.*\/)[^\/]+$/ ) {
            $real_dst = $1;
        }
        $real_dst.=$dst;
    }
    else {
        my @path_atoms = split( '/', $dst );
        $real_dst = $src;
        # truncate filename itself
        if( $real_dst =~ /(.+)\/[^\/]+$/ ) {
            $real_dst = $1;
        }

        my $i=0;
        while( $i < @path_atoms ) {
            if( $path_atoms[$i] eq '..' ) {
                if( $real_dst =~ /(.+)\/[^\/]+$/ ) {
                    $real_dst = $1;
                }
            }
            else {
                my $j=$i;
                while( $j < @path_atoms ) {
                    $real_dst.= "/".$path_atoms[$j];
                    $j++;
                }
                last;
            }
            $i++;
        }
    }

    return $real_dst;
}

# Check if the given file has the 'main' function
sub check_main_entry($ )
{
    my $filename = $_[0];
    my $filename_esc = escape_shell_symbols($filename);

    my $grep_main = `readelf -Wds $filename_esc | LC_ALL=C grep '\\ main\$\\|UND\\ __libc_start_main'`;
    chomp $grep_main;
    if( $grep_main ) {
        return 1;
    }

    return 0;
}

#--------------------------------------------------------------------------------
# Different exclusions from normal rules used by data collection and upload tools

# Commands sometimes reported by 'file' to be 'shared objects'
our %known_commands = ();
if( -e "known_commands" ) {
    open( CMDS, "known_commands" );
    while( <CMDS> ) {
        chomp;
        $known_commands{$_} = 1;
    }
    close( CMDS );
}
else {
    warn "Cannot open 'known_commands' file!";
}

# Files sometimes reported by 'file' as 'scripts' but which are not scripts, indeed
our %known_wrong_commands = (
    "CVS/Root" => 1,
    "CVS/Entries" => 1
);

# Symbols from libraries that are not actual interfaces that are used by apps
our %NotUploadableSymbols;
$NotUploadableSymbols{'_init'} = 1;
$NotUploadableSymbols{'_fini'} = 1;
$NotUploadableSymbols{'_DYNAMIC'} = 1;
$NotUploadableSymbols{'_GLOBAL_OFFSET_TABLE_'} = 1;

# Commands usually installed as /etc/alternatives or something like this
# Such commands are not visible when analyzing rpm or deb files, not the installed distribution
our %AliasedCommands;
$AliasedCommands{'lp.cups'} = 'lp';
$AliasedCommands{'lpr.cups'} = 'lpr';
$AliasedCommands{'lpq.cups'} = 'lpq';
$AliasedCommands{'lprm.cups'} = 'lprm';
$AliasedCommands{'lp-cups'} = 'lp';
$AliasedCommands{'lpr-cups'} = 'lpr';
$AliasedCommands{'lpq-cups'} = 'lpq';
$AliasedCommands{'lprm-cups'} = 'lprm';


1;
