#!/usr/bin/perl -w

# Linux Application Checker
# Application Analysis Data Generator (analysis.pl)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use strict;
use FindBin;
use File::Basename;

sub BEGIN { push @INC, $FindBin::Bin; }

use Analysis;
use Common;

my $time = 0;
sub elapsed {
    return "  ...".(time - $time)." sec\n";
}
sub trace {
	$time = time;
	return (@_, "...\n");
}

my ($std_version, $host_platform, $data_dir, $updata_file) = @ARGV;

if ( !defined $updata_file ) {
	die "Usage: $0 moblin_version host_platform updata_file";
}

my $res_dir = dirname($updata_file);
$res_dir = "." if $res_dir eq "";

my $arch_f = lc($host_platform);

## Load and analyze application data ##

sub find_data_file {
	my ($filename) = @_;

	my $file = $data_dir.'/'.$filename.'_'.$arch_f;
	if ( -f $file ) {
		return $file;
	}
	# If there is no unpacked file, try reading a gzipped file.
	if ( -f $file.'.gz' ) {
		return esc_cmdline("gzip -dc «$file.gz» |");
	}
	die "Can't read '$file'";
}

# Load distribution list (contains distribution name, version, database ID, architecture) 
print trace("load_distrs");
Analysis::load_distrs(find_data_file('Distributions'))
    or die "Failed to read the Distributions file.";
print elapsed();

# Read the already prepared information about the application components and their dependencies.
print trace("read_upload_data");
Analysis::read_upload_data($updata_file, $host_platform);
print elapsed();

# Work with libraries

print trace("read_Approved");
Analysis::read_Approved(find_data_file('Approved'));
print elapsed();

foreach my $com (values %$Analysis::app_components ) {
	Analysis::check_approved($com->{SONAME}) if defined $com->{SONAME};
	foreach my $soname ( @{$com->{NEEDED_LIBS}} ) {
		Analysis::check_approved($soname);
	}
}

# Read LibIndex file and filter only the needed information about libraries
# in different distributions.
print trace("read_LibIndex, read_UnapprovedLibs");
Analysis::read_LibIndex( find_data_file('LibIndex'), find_data_file('UnapprovedLibs') );
print elapsed();

# Read library dependencies.
print trace("read_RLibDeps");
my $lib_deps = Analysis::read_RLibDeps(find_data_file('RLibDeps'));
print elapsed();

# Read soname IDs used in library dependencies.
print trace("read_Sonames");
my $sonames = Analysis::read_Sonames(find_data_file('Sonames'));
print elapsed();

# Find a corresponding library in each distro for each required library.
print trace("link_libs");
Analysis::link_libs($lib_deps, $sonames);
print elapsed();

$lib_deps = undef; # No more needed
$sonames = undef; # No more needed

# Read STD history for required libraries.
print trace("read_STD_Libs");
Analysis::read_LSB_Libs(find_data_file('STD_Libs'));
print elapsed();

# Calculate the STD status for libraries.
print trace("std_lib_status");
Analysis::calc_lsb_lib_status($std_version);
print elapsed();

# Read additional info on required libraries.
print trace("read_LibAttrs");
Analysis::read_LibAttrs(find_data_file('STD_LibAttrs'));
print elapsed();

# Read other known compatibility issues.
print trace("read_libKB");
Analysis::read_libKB(find_data_file('LibKB'));
print elapsed();

# Work with interfaces.

# Read corresponding libraries on the local machine. This information will be
# used if we can't find in the database a corresponding library for some interface.
# It is assumed that the application under test is runnable on the local machine.
print trace("scan local libs");
Analysis::get_local_libs($host_platform);
print elapsed();

# Read parts of LibIfaces file where interface IDs of each required library
# (for each distribution) are listed.
print trace("find_libifaces");
Analysis::find_libifaces(find_data_file('LibIfaces'));
print elapsed();

$Analysis::distrlibs = undef; # No more needed

# Read index for the Interfaces_<ARCH> file.
# Keys in the index are calculated using the following formula:
# $key = substr( crc32($Iname), 0, 4) - 1000; # crc32 values are in range 1000..9999
# In the index file the line number is a key, the single integer in each line
# is an offset in the Interfaces_<ARCH> file.
print trace("read_IIndex");
my $int_index = Analysis::read_IIndex(find_data_file('IIndex'));
print elapsed();

# Find IDs for required interfaces. There could be several IDs for one interface name@ver.
print trace("find_interfaces");
my $riids = Analysis::find_interfaces(find_data_file('Interfaces'), $int_index);
print elapsed();
$int_index = undef; # No more needed

print trace("read_STD_Ints");
Analysis::read_LSB_Ints(find_data_file('STD_Ints'));
print elapsed();

# Find which library is an interface from for each required interface.
print trace("link_ints");
Analysis::link_ints($riids);
print elapsed();
$riids = undef; # No more needed

print trace("Improving interface linkage");
Analysis::improve_int_linkage($std_version);
print elapsed();

# Calculate the STD status for interfaces.
print trace("std_int_status");
Analysis::calc_lsb_int_status($std_version);
print elapsed();

print trace("read_IntAttrs");
Analysis::read_IntAttrs(find_data_file('STD_IntAttrs'));
print elapsed();

print trace("unmangle");
Analysis::unmangle();
print elapsed();

Analysis::complete_structures();

# TODO:
#Analysis::read_intKB(find_data_file('IntKB'));
#print elapsed("read_libKB");

## Save analysis results ##

print trace("save");
Analysis::save_main( $res_dir."/analysis_data" );
print elapsed();

# Copy the Distributions file
my $distr_file = $data_dir.'/Distributions_'.$arch_f;
if ( -f $distr_file ) {
	system(esc_cmdline("cp «$distr_file» «$res_dir/Distributions»"));
}
elsif ( -f $distr_file.'.gz' ) {
	system(esc_cmdline("gzip -dc «$distr_file.gz» >«$res_dir/Distributions»"));
}

# Remove the cached Summary (useful when re-analyzing the data)
system(esc_cmdline("rm «$res_dir/summary.htm» >/dev/null 2>&1"));

#-----------------------------------------------------------------------

# Calculate the overall verdict
Analysis::unneeded_libs();
my $verdict = Analysis::overall_verdict();

if (open(INFO, ">$res_dir/INFO")) {
	my $app_name = basename($updata_file);
	$app_name =~ s/^(.*)_\Q${host_platform}\E_upload_data/$1/;
	my $comment = read_file("$res_dir/test_logs/test_comments");
	print INFO <<DATA;
Application: $app_name
Verdict: $verdict

$comment
DATA
	close(INFO);
}
else {
	print STDERR "Failed to open INFO file for writing:\n$!\n";
}


exit 0;
