#!/usr/bin/perl -w
#################################################################################
# Prepare data about the given component to be inserted into the database.
#
# This script uses the output of the system package manager (rpm, dpkg,
# emerge or pacman), analyzing the files that are provided by the component.
#
# The script creates a text containing list of components, libraries and
# interfaces (actually for each shared object and executable the output of
# 'readelf' is placed into the output file; see README_collect for details).
# If the script is called for a standalone component then the output file should
# be processed by the 'upload_distr_data.pl' script on the server side for actual
# data insertion. In all other cases files produced by componenttodb.pl will
# be joined by apptodb.pl or distrtodb.pl and resultin files will be processed
# by appropriate upload* scripts.
#
# Classes will be detected while parsing interface names in the 'upload_data.pl'
#
# Symlinks and files that are not executable or shared objects are ignored.
#
# For shared objects, the script obtains ABItag using objdump
#
# Copyright (c) The Linux Foundation
# Copyright (c) Institute for System Programming, RAS
#
# Author: Denis Silakov <silakov@ispras.ru>
#################################################################################

use strict;
use Getopt::Long;
use File::Temp qw/ tempfile tempdir /;
use Text::Wrap;
Getopt::Long::Configure ("posix_default", "no_ignore_case");

use Distros;

# Component information
my $ComponentName;
my $ComponentVersion="";
my $ComponentId=0;
my $ComponentArch;
my $ComponentDbName;

# Package manager used in the distribution
my $PackageManager="";

# All about rpms
my @rpms;     # rpms that provide the component
my @sofiles;  # list of files from rpms that provide the component

# All about library under process
my $LibName;
my $LibRunName = "";
my $LibSoName = "";
my $FileName = "";
my $LibPath = "";
my $LibPath_p = "";

# Options
my $opt_help;
my $opt_java=0;
my $OutputFile;
my $ComponentFilesList;
my $quiet=1;
my $totally_quiet;
my $analyze_il=1; # Should we analyze interpreter language modules used by scripts and other modules
my $analyze_cmds=1; # Should we analyze dependencies of executable files
my $trsp_symlink;

my %path_subst; # Replacements for temporary filenames

# Max number of forks when processing java components
my $java_max_jobs=3;

GetOptions("help|h!" => \$opt_help,
            "verbose" => sub {$quiet=0;},
            "quiet|q!" => \$totally_quiet,
            "arch|a=s" => \$ComponentArch,
            "java|j!" => \$opt_java,
            "distromode!" => sub {$analyze_il=0; $analyze_cmds=0},
            "packagemanager|m=s" => \$PackageManager,
            "comp-dbname|c=s" => \$ComponentDbName,
            "version|v=s" => \$ComponentVersion,
            "filelist|l=s" => \$ComponentFilesList,
            "file|f=s" => \$OutputFile,
            "trsp-symlink" => sub {$trsp_symlink=1;},
            );

sub HELP_MESSAGE();
sub add_component();
sub add_library();

# Functions for obtaining information using different package managers
sub get_info_rpm();
sub get_info_dpkg();
sub get_info_emerge();
sub get_info_pacman();
sub get_info_by_filelist();

sub process_perl_module($$ );

HELP_MESSAGE() if (defined $opt_help);

if(@ARGV != 1) {HELP_MESSAGE(); exit(1);}

my ($th, $tmpFile) = tempfile(UNLINK => 1);
$tmpFile = escape_shell_symbols($tmpFile);

# This directory is used to process jar files contents.
# All files in it are deleted every time the 'jar' file is met,
# so don't use it for other purposes
my $tmpDir = tempdir( CLEANUP => 1 );
$tmpDir = escape_shell_symbols($tmpDir);

# Command used to analyze java classes
my $parse_java_cmdline = "java -jar java/dx.jar --dump --width=1000";
my $java_analysis_postprocess = 'perl -ne "if( /^constant_pool/ .. /^end constant_pool/ ) {} else {print \$_;}"';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: [ila]load\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: goto\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: ifnull\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: ifeq\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: [ial]const\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: [ila]store\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: [li]a[dn]d\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: dup\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: if[lgn][et]\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: if_[ali]cmp\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: [li]sub\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: [ali]return\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: ish[rl]\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: swap\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: iinc\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: [libca]astore\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: return\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: [sb]ipush\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: imul\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: pop\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: ifnonnull\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: arraylength\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'^[[:space:]]*[[:xdigit:]]\{4\} [[:xdigit:]]\{2,4\}\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: ldc\'';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: athrow\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: [abilc]aload\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: newarray\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: anewarray\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: i2[lb]\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: checkcast\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: monitor\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}\.\.[:xdigit:]]\{4\}\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: tableswitch\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: l2i\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: ixor\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: lookupswitch\' ';
$java_analysis_postprocess.= '| LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: instanceof\' | LC_ALL=C grep -v \'[[:xdigit:]]\{4\}: lookupswitch\' ';

# Guess package manager, if it was not specified
if( (not $PackageManager) and (not $ComponentFilesList) ) {
    $PackageManager = guess_package_manager();
    if( $PackageManager eq "unknown" ) {
        print STDERR "Cannot guess package manager! Please specify it directly using '-p' option.";
        exit(1);
    }
    else {
        print STDERR "No package manager was specified, guessing it is $PackageManager\n";
    }
}

$ComponentName = $ARGV[0];

if( not $ComponentDbName ) {
    $ComponentDbName = $ComponentName;
}

if( !defined $OutputFile ) {
    $OutputFile = $ComponentName."_upload_data";
    $OutputFile =~ s/\//_/g;
}

if( -e $OutputFile ) {
    print STDERR "!!! Can't dump component data - file '$OutputFile' already exists\n";
    print STDERR "!!! Please remove or rename the file first, or specify another output file using '-o' option.\n";
    exit(1);
}

my $OutputFile_esc = escape_shell_symbols($OutputFile);

open( COMPONENT_DATA, "> $OutputFile" );

if( not $ComponentArch ) {
    $ComponentArch = guess_platform(`uname -m`);
}

##############################################################################
# Main section
##############################################################################

if( $ComponentFilesList ) {
    get_info_by_filelist();
}
elsif( $PackageManager eq "rpm" ) {
    get_info_rpm();
}
elsif( $PackageManager eq "dpkg" ) {
    get_info_dpkg();
}
elsif( $PackageManager eq "pacman" ) {
    get_info_pacman();
}
elsif( $PackageManager eq "emerge" ) {
    if( does_shell_know("equery") ) {
        get_info_emerge();
    }
    else {
        die "This script requires 'equery' tool from 'gentoolkit' package";
    }
}
else {
    die "Package manager '$PackageManager' is not supproted";
}

my $python_email_processed=0;
my $python_gtk_processed=0;

# Check the types of every file and call appropriate function
foreach my $soname ( @sofiles )
{
    if( !$soname ) {
        next; # we can get empty strings from dpkg, ignore them
    }
    print "Processing file: '$soname'\n" if(not $quiet);
    my $filename = escape_shell_symbols($soname);

    # What does 'file' think about our file - is it an executable or a shared object?
    # Use brief mode - the filename can contain words that will influence analysis.
    my $type = `file -b $filename`;

    my $basename = `basename $filename`;
    chomp($basename);

    my $comment = "";

    my $soname_p;
    if ( exists $path_subst{$soname} ) {
        $soname_p = $path_subst{$soname};
    } else {
        $soname_p = $soname;
    }

    # Process symlinks
    if( -l($soname) ) {
        my $target_name = "";
        my $target_name_esc = "";
        my $target_shortname = "";
        my $current_name = $soname;
        do { # resolve the chain of symlinks
            if( $type =~ /link to `(.+)'/ ) {
                $target_name = $1;

                # Relative path?
                if( $target_name !~ /^\// ) {
                    $target_name = calc_abs_path( $current_name, $target_name );
                    $current_name = $target_name;
                }

                $target_shortname = $target_name;
                if( $target_name !~ /\// ) {
                    my $target_base = $soname;
                    $target_base =~ s/\Q$basename\E$/\Q$target_name\E/;
                    $target_name = $target_base;
                }

                $target_name_esc = escape_shell_symbols($target_name);
                $type = `file -b $target_name_esc`;
            }
            else {
                last;
            }
        } while ($type =~ /link to `(.+)'/);

        $target_name_esc = escape_shell_symbols($target_name);
        $target_shortname = `basename $target_name_esc`;
        chomp($target_shortname);

        my $main_entry_present = 0;
        if( $type =~ /shared object/i and $target_shortname !~ /^lib.+\.so.*/ ) {
            $main_entry_present = check_main_entry($target_name);
        }

        #~ if( $type !~ /executable/i and !$trsp_symlink and !defined($known_commands{$target_shortname})
        if( $type !~ /executable/i and !$trsp_symlink and !$main_entry_present
                and ($type !~ /No such file/ or $target_name !~ /bin\//) ) {
            print "    Symlink to not-executable, skipping...\n" if(not $quiet);
            next;
        }

        #~ if( defined($known_commands{$target_shortname}) or $type =~ /No such file/  ) {
        if( $main_entry_present or $type =~ /No such file/  ) {
            $type = 'executable';
        }

        if( !$quiet and $type =~ /No such file/ ) {
            print "    Treating broken symbolic link as command since its target has 'bin/' in the name.\n";
        }

        $comment = "Symlink to $target_name";
    }

    if( $type =~ m/(a )?(\w(\s|\:|\-|\/|\w|\+|\.)*)\s+script/ ) {
        if( defined( $known_wrong_commands{$basename} ) ) {
            # Skip files sometimes reported as commands/scripts, but which are known to be not commands or scripts
            next;
        }
        print "Found a script: $soname (type: '$type')\n" if(not $quiet);
        print COMPONENT_DATA "!Script '$basename' '$soname_p' '$2'\n";
        my $interpreter = `head -n 1 $filename`;
        chomp( $interpreter );
        if( $interpreter !~ /^\#\!/ ) {
            print STDERR "    Can't detect required interpreter - the script doesn't start with '#!'\n" if( !$quiet );
            $interpreter = "";
        }
        else {
            $interpreter =~ s/^\#\!//;
        }

        if( $type =~ /perl\s+script/i or $interpreter =~ /perl/ ) {
            print COMPONENT_DATA "  PerlInterpreter: $interpreter\n";
            if( $analyze_il ) {
                close( COMPONENT_DATA );
                system( "./interpreted\\ languages/analyze_perl.pl $filename >> $OutputFile_esc" ) == 0 or die();
                open( COMPONENT_DATA, ">> $OutputFile" );
            }
        }
        elsif( $type =~ /python\s+script/i or $interpreter =~ /python/ ) {
            print COMPONENT_DATA "  PythonInterpreter: $interpreter\n";
            if( $analyze_il ) {
                close( COMPONENT_DATA );
                system( "./interpreted\\ languages/analyze_python.py $filename | LC_ALL=C sort -u >> $OutputFile_esc" ) == 0 or die();
                open( COMPONENT_DATA, ">> $OutputFile" );
            }
        }
        elsif( $soname =~ /\.pm$/ and $type =~ /text/ ) {
            print "Found a perl module1: $soname\n" if(not $quiet);
            process_perl_module( $filename, $soname_p );
            next;
        }
        elsif( $soname =~ /\.pyc?$/i ) {
            # This is mostly empiric check - even if file thinks that we don't have a python script,
            # but the file ends with 'py', let's try to analyze it and add as a module if succeeded
            my $pythonResult = `./interpreted\\ languages/analyze_python.py $filename`;
            if( $pythonResult !~ /Cannot analyse/ ) {
            # See comments below for such checks
#               my $modname = $basename;
#               $modname =~ s/\.pyc?$//i;
#               if( $modname eq "__builtin__" or $modname eq "__main__" or $modname eq "__init__" or $modname eq "__future__" ) {
#                   next;
#               }
#               my $py_import_path = $soname;
#               $py_import_path =~ s/\Q\/$basename\E//;
#               my $errors = "";
#               if( $quiet ) {
#                   $errors = "2>/dev/null";
#               }
#
#               my $exports_str = `PYTHONPATH=\$PYTHONPATH:$py_import_path python -c "import sys; import os; import $modname; print os._get_exports_list($modname)" $errors | sed s/\\\\\\[// | sed s/\\\\\\]//| sed s/\\'//g | sed s/\\ //g`;
#               chomp( $exports_str );
#               my @exports = split( ',', $exports_str );
#               my $count = 0;
#               while( $count <= $#exports ) {
#                   my $expname = $exports[$count];
#                   print COMPONENT_DATA "!PythonModule '$modname.$expname' '$soname_p' 'python'\n";
#                   $count++;
#               }
                if( $interpreter =~ /^\#\!/ ) {
                    $interpreter =~ s/^\#\!//;
                    print COMPONENT_DATA "!Script '$basename' '$soname_p' 'python'\n";
                }
                else {
                    $interpreter = "";
                    print COMPONENT_DATA "!PythonModule '$basename' '$soname_p' 'python'\n";
                }
                print COMPONENT_DATA "  PythonInterpreter: $interpreter\n";
                if( $analyze_il ) {
                    print COMPONENT_DATA $pythonResult;
                }

            }
        }
        else {
            print COMPONENT_DATA "  ScriptInterpreter: $interpreter\n";
        }

        next;
    }
    elsif( $type =~ /perl.+module/i or ($soname =~ /\.pm$/ and $type =~ /text/ ) ) {
        print "Found a perl module: $soname\n" if(not $quiet);
        process_perl_module( $filename, $soname_p );
        next;
    }
    elsif( $type =~ /DOS executable/ ) {
        print "Found MS-DOS executable: $soname\n" if(not $quiet);
        print COMPONENT_DATA "!Command '$basename' '$soname_p'\n";
        next;
    }
    elsif( $type =~ m/executable/ ) {
        if( defined( $known_wrong_commands{$basename} ) ) {
            # Skip files sometimes reported as commands, but which are known to be not commands
            next;
        }
        print "Found a command: $soname\n" if(not $quiet);
        if( $soname =~ /\.py$/ ) {
            print COMPONENT_DATA "!PythonModule '$basename' '$soname_p' 'python'\n";
        }
        print COMPONENT_DATA "!Command '$basename' '$soname_p' '$comment'\n";
        if( $comment =~ /Symlink/ and !$trsp_symlink ) {
            next;
        }

        if( $analyze_cmds ) {
            print "Calling readelf...\n" if(not $quiet);
            my $cmd = "LC_ALL=C readelf -WhSsdA $filename | LC_ALL=C grep -v NOTYPE | LC_ALL=C grep -Fv 'SECTION LOCAL  DEFAULT' | LC_ALL=C grep -Fv 'LOCAL  HIDDEN' | perl -ne \"if(/Symbol table '.symtab'/.../^\\S/ ) { print \\\$_ if( /^\\S/ and ! /Symbol table '.symtab'/ );} else { print \\\$_; }\" > $tmpFile";
            system( $cmd ) == 0 or die();
            close( COMPONENT_DATA );
            system "cat $tmpFile >> $OutputFile_esc";
            open( COMPONENT_DATA, ">> $OutputFile" );
        }
        next;
    }
    elsif( $type =~ m/shared object/ ) {
        # Sometimes commands are also reported to be 'shared objects' by file...
        # Let's check our special arrays...
        #if( defined( $known_commands{$basename} ) ) {
        if( $basename !~ /^lib.+\.so.*/ and check_main_entry($soname) ) {
            print "Found a command: $soname\n" if(not $quiet);
            print COMPONENT_DATA "!Command '$basename' '$soname_p'\n";

            if( $analyze_cmds ) {
                print "Calling readelf...\n" if(not $quiet);
                my $cmd = "LC_ALL=C readelf -WhSsdA $filename | LC_ALL=C grep -Fv NOTYPE | LC_ALL=C grep -Fv 'SECTION LOCAL  DEFAULT' | LC_ALL=C grep -Fv 'LOCAL  HIDDEN' | perl -ne \"if(/Symbol table '.symtab'/.../^\\S/ ) { print \\\$_ if( /^\\S/ and ! /Symbol table '.symtab'/ );} else { print \\\$_; }\" > $tmpFile";
                system( $cmd ) == 0 or die();
                close( COMPONENT_DATA );
                system "cat $tmpFile >> $OutputFile_esc";
                open( COMPONENT_DATA, ">> $OutputFile" );
            }
            next;
        }

        print "Found a shared object: $soname\n" if(not $quiet);

        $LibRunName = $basename;

        $LibPath = $soname;
        $LibPath =~ s/\/\Q$basename\E$//;

        $LibPath_p = $soname_p;
        $LibPath_p =~ s/\/[^\/]*$//;

        $FileName = $soname;
        ($LibName, my @temp) = split /\.so/, $LibRunName;

        # Add library data
        add_library();

        print "Calling readelf...\n" if(not $quiet);
        my $cmd = "LC_ALL=C readelf -WhSsdA $filename | LC_ALL=C grep -Fv NOTYPE | LC_ALL=C grep -Fv 'SECTION LOCAL  DEFAULT' | LC_ALL=C grep -Fv 'LOCAL  HIDDEN' | perl -ne \"if(/Symbol table '.symtab'/.../^\\S/ ) { print \\\$_ if( /^\\S/ and ! /Symbol table '.symtab'/ );} else { print \\\$_; }\" > $tmpFile";
        system( $cmd ) == 0 or die() ;
        close( COMPONENT_DATA );
        system "cat $tmpFile >> $OutputFile_esc";
        open( COMPONENT_DATA, ">> $OutputFile" );
    }
    elsif( $soname =~ /\.pyc?$/i ) {
        my $interpreter = `head -n 1 $filename`;
        chomp( $interpreter );
        my $pythonResult = `./interpreted\\ languages/analyze_python.py $filename`;
        if( $pythonResult !~ /Cannot analyse/ ) {
            my $modname = $basename;
            $modname =~ s/\.pyc?$//i;
            if( $modname eq "__builtin__" or $modname eq "__main__" or $modname eq "__init__" or $modname eq "__future__" ) {
                next;
            }

            # !!! Next checks are too rough. First, we'll obtain too many functions;
            #     second, importing modules can be dangerous. So do some primitive checks
            #     for 'os' exports only
            if( $modname eq "os" ) {
                my $exports_str = `python -c "import sys; import os; print os._get_exports_list(os)" 2>/dev/null | sed s/\\\\\\[// | sed s/\\\\\\]//| sed s/\\'//g | sed s/\\ //g`;
                foreach my $submod_required ('altsep', 'curdir', 'pardir', 'sep', 'pathsep', 'linesep', 'defpath', 'name', 'path', 'devnull') {
                    if( $exports_str =~ /\Q$submod_required\E/ ) {
                        print COMPONENT_DATA "!PythonModule 'os.$submod_required' '$soname_p' 'python'\n";
                    }
                }
            }
            elsif( $soname =~ /email/ and !$python_email_processed ) {
                my $exports_str = `python -c "import sys; import os; import email; print os._get_exports_list(email)" 2>/dev/null | sed s/\\\\\\[// | sed s/\\\\\\]//| sed s/\\'//g | sed s/\\ //g`;
                foreach my $submod_required ('base64MIME', 'Charset', 'Encoders', 'Errors', 'Generator', 'Header',
                                'Iterators', 'Message', 'MIMEAudio', 'MIMEBase', 'MIMEImage', 'MIMEMessage', 'MIMEMultipart',
                                'MIMENonMultipart', 'MIMEText', 'Parser', 'quopriMIME', 'Utils') {
                    if( $exports_str =~ /\Q$submod_required\E/ ) {
                        print COMPONENT_DATA "!PythonModule 'email.$submod_required' '$soname_p' 'python'\n";
                    }
                }
                $python_email_processed = 1;
            }
            elsif( $soname =~ /\/gtk\// and !$python_gtk_processed ) {
                my $exports_str = `python -c "import sys; import os; import gtk; print os._get_exports_list(gtk)" 2>/dev/null | sed s/\\\\\\[// | sed s/\\\\\\]//| sed s/\\'//g | sed s/\\ //g`;
                foreach my $submod_required ('gdk') {
                    if( $exports_str =~ /\Q$submod_required\E/ ) {
                        print COMPONENT_DATA "!PythonModule 'gtk.$submod_required' '$soname_p' 'python'\n";
                    }
                }
                $python_gtk_processed = 1;
            }

            # Try to detect what does the module provide, since some of its exports might be
            #  required by apps directly
#           my $py_import_path = $soname;
#           $py_import_path =~ s/\Q\/$basename\E//;
#
#           my $errors = "";
#           if( $quiet ) {
#               $errors = "2>/dev/null";
#           }
#
#           my $exports_str = `PYTHONPATH=\$PYTHONPATH:$py_import_path python -c "import sys; import os; import $modname; print os._get_exports_list($modname)" $errors | sed s/\\\\\\[// | sed s/\\\\\\]//| sed s/\\'//g | sed s/\\ //g`;
#           chomp( $exports_str );
# #             print "$soname exploited to $exports_str\n";
#           my @exports = split( ',', $exports_str );
#           my $count = 0;
#           while( $count <= $#exports ) {
#               my $expname = $exports[$count];
#               print COMPONENT_DATA "!PythonModule '$modname.$expname' '$soname_p' 'python'\n";
#               $count++;
#           }
#
            if( $interpreter =~ /^\#\!/ ) {
                $interpreter =~ s/^\#\!//;
                print COMPONENT_DATA "!Script '$basename' '$soname_p' 'python'\n";
            }
            else {
                $interpreter = "";
                print COMPONENT_DATA "!PythonModule '$basename' '$soname_p' 'python'\n";
            }
            print COMPONENT_DATA "  PythonInterpreter: $interpreter\n";
            if( $analyze_il ) {
                print COMPONENT_DATA $pythonResult;
            }
        }
    }
    elsif( $type =~ /data/ and $soname =~ /\.class$/ and $opt_java ) { # `file` is not familar with java classes
        system( "$parse_java_cmdline $filename | $java_analysis_postprocess >$tmpFile" ) == 0 or die();
        my $isException = `head -3 $tmpFile | tail -1`;
        if( $isException !~ /UNEXPECTED TOP-LEVEL EXCEPTION/ ) {
            print COMPONENT_DATA "!JavaClassFile '$basename' '$soname_p'\n";
            close COMPONENT_DATA;
            system "cat $tmpFile >> $OutputFile_esc";
            open( COMPONENT_DATA, ">> $OutputFile" );
        }
        elsif( not $quiet ) {
            print "    File has '.class' suffix but doesn't seem to be Java class...\n";
        }
    }
    elsif( $type =~ /Zip archive data/ and $soname_p =~ /\.jar$/ and $opt_java ) {
        print "    Jar archive detected\n" if not $quiet;
        my $basename_esc = escape_shell_symbols($basename);
        if( !does_shell_know('jar') ) {
            if( !does_shell_know('unzip') ) {
                warn "You need 'jar' or 'unzip' command to analyze jar archives.";
                next;
            }

            system "rm -rf $tmpDir/*";
            my $zip_options = $quiet ? "-qq" : "";
            system( "unzip $zip_options -d $tmpDir $filename" ) == 0 or die();
        }
        else {
            system "rm -rf $tmpDir/*";
            system "cp $filename $tmpDir && cd $tmpDir && jar -xf $basename_esc && rm -f $basename_esc";
        }
        open( JAR_CONTENTS, "find $tmpDir -type f -name \"*.class\" -print |" );
        my $jobs = 0;

        while( <JAR_CONTENTS> ) {
            $jobs++;
            defined (my $pid = fork) or die "Can't fork: $!";
            if( $pid == 0 ) {
                chomp;
                # Let's hope that all .class files inside jar are really Java class files.
                # Finally, there is no harm if they are not - we'll simply see exception from analyzer in the data file
                my $classFileName = escape_shell_symbols($_);
                my $classDataFile = $classFileName."_data";
                system "echo \"!JavaClassFile '$basename' '$classFileName'\" >$classDataFile";
                system( "$parse_java_cmdline $classFileName | $java_analysis_postprocess >>$classDataFile" ) == 0 or die();
                exit;
            }
            elsif( $jobs >= $java_max_jobs ) {
                wait;
                $jobs--;
            }
        }

        my $child_pid = wait;
        while( $child_pid >= 0 ) {
            $child_pid = wait;
        }

        close( JAR_CONTENTS );

        close COMPONENT_DATA;
        system( "find $tmpDir -type f -name \"*.class_data\" -exec cat \{\} >>$OutputFile_esc \\;" ) == 0 or die();
        open( COMPONENT_DATA, ">> $OutputFile" );
    }
    else {
        print "    Not executable or shared object, not a perl/python script/object, skipping...\n" if( not $quiet );
    }
}

# Delete "$tmpDir" from result file if we are processing rpms or debs
my $tmpDir_escaped = $tmpDir;
$tmpDir_escaped =~ s/\//\\\\\//g;
system "LC_ALL=C sed -i s/$tmpDir_escaped//g $OutputFile_esc";

################################################################################################
# Collect information about components directly by file list or using different package managers
################################################################################################

sub get_info_by_filelist()
{
    open( FILE_LIST, "< $ComponentFilesList" ) || die "Cannot open file with filelist ($ComponentFilesList)";

    my $tmpName = $ComponentName;
    if( $ComponentName =~ /rpm$/ ) {
        $ComponentName =~ s/\/.*\///g;
        $ComponentName =~ s/\-[0-9].*//g;
        $ComponentDbName = $ComponentName;
    }
    elsif( $ComponentName =~ /deb$/ ) {
        $ComponentName =~ s/\/.*\///g;
        $ComponentName =~ s/\_.*//g;
        $ComponentDbName = $ComponentName;
    }
    else {
        $ComponentName = $ComponentDbName;
    }
    add_component();
    $ComponentName = $tmpName;

    my $type;

    while( my $filename = <FILE_LIST> ) {
        chomp($filename);
        my $subst = undef;
        # There could be specified a replacement for path/filename
        # that should be printed in the output file.
        # (usually to replace temp names with smth more useful).
        # It goes after first '|'
        if ( $filename =~ /^([^\|]*)\|(.*)$/ ) { # look for substitution
            $subst = $filename;
            $filename = $1;
        }

        if( !-e $filename ) {
            warn "File '$filename' from the list doesn't exist!" if(!$totally_quiet);
            next;
        }

        my $filename_esc = escape_shell_symbols($filename);
        $type = `file -b $filename_esc`;

        if( $type =~ /directory/ ) {
            local *find_files = sub {
                my ($cmd) = @_;

                open( FILES, "$cmd |" )
                    or die "Cannot process files from directory $_";

                while( my $found = <FILES> ) {
                    chomp $found;
                    if ( defined $subst ) {
                        my $new_subst = $found;
                        $new_subst =~ s/^\Q$filename\E/\Q$subst\E/;
                        $path_subst{$found} = $new_subst;
                    }
                    push @sofiles, $found;
                }
                close FILES;
            };

            &find_files( "find $filename_esc -type f -print" );

            # Process symlinks, too
            &find_files( "find $filename_esc -type l -print" );
        }
        else {
            push @sofiles, $filename;
            if ( defined $subst ) {
                $path_subst{$filename} = $subst;
            }
        }
    }

    close FILE_LIST;
}

sub get_info_rpm()
{
    my $ComponentName_esc = escape_shell_symbols($ComponentName);

    # Get a list of rpm that provide the given component
    open(RPM_PROV,"rpm -q --whatprovides $ComponentName_esc|") or die "Cannot get information from rpm.";
    @rpms = <RPM_PROV>;

    chomp($rpms[0]);
    if( $rpms[0] =~ /no package provides \Q$ComponentName\E/ ) {
        print STDERR "$ComponentName: can't get information\n Your package manager doesn't know anything about '$ComponentName' package, \n"
                . " neither file with this name exists\n";
    }

    close RPM_PROV;

    # If component version is not set manually, then take it from the first rpm
    if( !$ComponentVersion ) {
        if ($rpms[0] =~ /((\w|\d|\-\w)*)\-((\w|\d|\.|\-)*)/ ) {
            $ComponentVersion = $3;
        }
        else {
            print STDERR "$ComponentName: Cannot get component version!\n";
            $ComponentVersion = "";
        }
    }

    add_component();

    # Some rpm versions know only '--list', the others - 'filesbypkg' or 'fileprovide'...
    # Let's guess
    my $request = '--list';
    my $try = `rpm --help | grep list`;
    if( !$try ) {
        $try = `rpm --help | grep fileprovide`;
        if( !$try ) {
            $try = `rpm --help | grep filesbypkg`;
            if( !$try ) {
                die "Cannot guess how to extract package file list from rpm!";
            }
            else {
                $request = '--filesbypkg';
            }
        }
        else {
            $request = '--fileprovide';
        }
    }

    # Explore files provided by rmps
    for( 0..$#rpms ) {
        chomp($rpms[$_]);

        open(FILE_PROV,"rpm -q $request ".escape_shell_symbols($rpms[$_])." | ") or die "Cannot get information from rpm.";

        while(<FILE_PROV>) {
            chomp;

            if( /not installed / ) {
                warn "Missing package!";
                next;
            }

            if( /((\/(\w|\.|\-|\+|\[)+)*)\s*((\w+\/\s*)+)*/ ) {
                print "file found: '".$1."'\n" if(not $quiet);
                push @sofiles, $1;
            }
        }

        close FILE_PROV;
    }
}

sub get_info_dpkg()
{
    # Unfortunately, we cannot use something like 'rpm -q --whatprovides' here,
    # cause for many base components (like Qt, Gtk etc.) there are no
    # packages that 'Provide' them. So, user should provide us with correct
    # package name

    my $ComponentName_esc = escape_shell_symbols($ComponentName);

    # Check if the component is available
    my $notAvailable = `dpkg -s $ComponentName_esc | grep "is not installed"`;
    if( $notAvailable ) {
        warn "Package $ComponentName is not installed and cannot be processed.";
        return;
    }

    # If component version is not set manually, then take it from the first package
    if( !$ComponentVersion ) {
        my $ver = `dpkg -s $ComponentName_esc | LC_ALL=C grep -i version`;
        if( $ver =~ /(V|v)ersion\s*:\s*(\S+)/ ) {
            $ComponentVersion = $2;
        }
        else {
            warn "$ComponentName: Cannot get component version!";
            $ComponentVersion = "";
        }
    }

    add_component();

    open(FILE_PROV,"dpkg -L $ComponentName_esc | ") || die "Cannot get information from dpkg.";
    while(<FILE_PROV>) {
        chomp;
        print "file found: '".$_."'\n" if(not $quiet);
        push @sofiles, $_;
    }

    close FILE_PROV;
}

sub get_info_emerge()
{
    my $ComponentName_esc = escape_shell_symbols($ComponentName);

    # Get component version, if it is not specified directly
    if( !$ComponentVersion ) {
        my $VerLine = `equery which $ComponentName_esc`;
        if( $VerLine =~ /\Q$ComponentName\E-(.+)\.ebuild/ ) {
            $ComponentVersion=$1;
        }
        else {
            warn "Cannot get component version for '$ComponentName'!";
        }
    }

    add_component();

    open(FILE_PROV,"equery files $ComponentName_esc | ") || die "Cannot get information from equery.";
    while(<FILE_PROV>) {
        chomp;
        print "file found: '".$_."'\n" if(not $quiet);
        push @sofiles, $_;
    }

    close FILE_PROV;
}

sub get_info_pacman()
{
    my $ComponentName_esc = escape_shell_symbols($ComponentName);

    # Get component version, if it is not specified directly
    if( !$ComponentVersion ) {
        my $VerLine = `pacman -Qi $ComponentName_esc | LC_ALL=C grep Version | cut -f2- -d:`;
        chomp $VerLine;
        if( $VerLine =~ /\s*(.+)\s*$/ ) {
            $ComponentVersion = $1;
        }
        else {
            warn "Cannot get component version for '$ComponentName'!";
        }
    }

    add_component();

    open(FILE_PROV,"pacman -Ql $ComponentName_esc | grep -v '/\$' | cut -d\\  -f2- |") || die "Cannot get information from pacman.";
    while(<FILE_PROV>) {
        chomp;
        print "file found: '".$_."'\n" if(not $quiet);
        push @sofiles, $_;
    }

    close FILE_PROV;
}

##############################################################################
# Functions for printing additional information about artifacts
##############################################################################

# Utility function for processing of objdump output.
sub parse_dword {
    my ( $dword, $byte_order ) = @_;

    if ( $byte_order == 0 ) { # little-endian
        my $reversed = "";
        while ( $dword =~ s/^(\d{2})// ) {
            $reversed = $1.$reversed;
        }
        $dword = $reversed;
    }

    return hex($dword);
}

#
# add_library: print record with library information
#
sub add_library()
{
    # Get soname of the library
    $LibSoName='';
    my $SonameString='';
    my $search_string = escape_shell_symbols("$LibPath/$LibRunName");

    $SonameString = `LC_ALL=C readelf -d $search_string | LC_ALL=C grep -F SONAME`;
    if( $SonameString =~ /Library soname: \[((\S)*)\]/ ) {
        $LibSoName = $1;
    }

    # Get ABI-tag
    open OBJDUMP, "objdump -s -j .note.ABI-tag $search_string |"
            or die "Cannot get information from objdump.";

    my @dwords = (); # Contents of '.note.ABI-tag' section
                     # in 32 bit dwords in a hexadecimal representation
                     # with a platform-specific byte order.

    while ( my $line = <OBJDUMP> ) {
        chomp $line;

        if ( $line =~ /^ \d+/ ) { # Contents of the section .note.ABI-tag
            #Format:  offset dword_a{8} dword_b{8} dword_b{8} dword_d{8}  ...ASCII...
            if ( $line =~ /^ [\dA-F]+\s+([\dA-F]{8})\s+([\dA-F]{8})\s+([\dA-F]{8})\s+([\dA-F]{8})\s+/i ) {
                push @dwords, $1, $2, $3, $4;
            }
            else {
                die "Can't parse objdump output: '$line'";
            }
        }
    }

    close( OBJDUMP );

    my $ABItag = "";

    # Process the obtained data
    if ( @dwords ) {
        scalar(@dwords) >= 8 or die "Too little bytes in the section.";

        # See LSB-Core-generic/noteabitag.html

        my $byte_order = 0; # little-endian
        # Guess the byte order. Not the best way, but just works.
        if ( $dwords[2] eq "00000001" ) {
            $byte_order = 1; # big endian
        }

        parse_dword( $dwords[0], $byte_order ) == 4
            or die "The first dword ('namesz' field?) should equals 4";
        parse_dword( $dwords[1], $byte_order ) == 16
            or die "The second dword ('descsz' field) should equals 16";
        parse_dword( $dwords[2], $byte_order ) == 1
            or die "The third dword ('type' field) should equals 1";
        $dwords[3] eq "474e5500"
            or die "The fourth dword ('name' field) should be \"GNU\\0\"";
        parse_dword( $dwords[4], $byte_order ) == 0
            or die "The fifth dword (1st dword in 'desc' field) should be 0 (this signifies a Linux executable)";

        my $a = parse_dword( $dwords[5], $byte_order );
        my $b = parse_dword( $dwords[6], $byte_order );
        my $c = parse_dword( $dwords[7], $byte_order );

        $ABItag = "Linux $a.$b.$c";
    }

    # Not perfect, but we need some way to distinct python modules provided as *.so files
    if( $LibPath =~ /python/ ) {
        print COMPONENT_DATA "!PythonModule '$LibName' '$LibPath_p/$LibName' 'python'\n";
    }

    # Try to detect library version (quite difficult task...)
    if( ($LibName =~ /^ISO/) or ($LibName =~ /^GOST/) or ($LibName =~ /^UTF/) or ($LibName =~ /^TIS/)
            or ($LibName =~ /^ISIRI/) ) {
        print COMPONENT_DATA "!Library '$LibName' '' '$LibRunName' '$LibPath_p' '$ABItag' '$LibSoName'\n";
    }
    elsif( $LibName =~ /(lib(g|stdc)\+\+\-(\d-)?libc)(\d\..+)/ ) {
        my $lname = $1;
        my $lver = $4;
        print COMPONENT_DATA "!Library '$lname' '$lver' '$LibRunName' '$LibPath_p' '$ABItag' '$LibSoName'\n";
    }
    elsif( $LibName =~ /((\w|\d|_|\+|\-\w)+)\-((\d).+)/ ) {
        my $lname = $1;
        my $lver = $3;
        print COMPONENT_DATA "!Library '$lname' '$lver' '$LibRunName' '$LibPath_p' '$ABItag' '$LibSoName'\n";
    }
    elsif( $LibRunName =~ /((\w|\d|_|\+|\.|\-\w)+)\.so\.((\d).+)/ ) {
        my $lname = $1;
        my $lver = $3;
        print COMPONENT_DATA "!Library '$lname' '$lver' '$LibRunName' '$LibPath_p' '$ABItag' '$LibSoName'\n";
    }
    else {
        print COMPONENT_DATA "!Library '$LibName' '' '$LibRunName' '$LibPath_p' '$ABItag' '$LibSoName'\n";
    }
}

sub add_component()
{
    # Remove the leading '=', if presents, in case of Gentoo portage
    if( $PackageManager eq "emerge" ) {
        my $PackageName = $ComponentName;
        $PackageName =~ s/^\=//;
        if( $PackageName =~ /(-\Q$ComponentVersion\E)$/ ) {
                $PackageName =~ s/-\Q$ComponentVersion\E//;
        }
        print COMPONENT_DATA "!Component '$PackageName' '$ComponentVersion' '$ComponentDbName' '$ComponentArch'\n";
    }
    else {
        print COMPONENT_DATA "!Component '$ComponentName' '$ComponentVersion' '$ComponentDbName' '$ComponentArch'\n";
    }
}

##############################################################################
# Functions to work with interpreted languages
##############################################################################
sub process_perl_module($$ )
{
    my $filename = $_[0];
    my $soname_p = $_[1];

    open( PERL_PROV, "echo $filename | ./interpreted\\ languages/perl.prov | " );
    while( <PERL_PROV> ) {
        chomp;
        if( /perl\((.+)\)( = (.+))?/ ) {
            my $ver = defined $3 ? $3 : '';
            print COMPONENT_DATA "!PerlModuleProvided '$1' '$soname_p' '$ver'\n";
        }
    }
    close( PERL_PROV );
    if( $analyze_il ) {
        close( COMPONENT_DATA );
        system( "./interpreted\\ languages/analyze_perl.pl $filename >> $OutputFile_esc" ) == 0 or die();
        open( COMPONENT_DATA, ">> $OutputFile" );
    }
}

##############################################################################
# Miscellaneous functions
##############################################################################

sub HELP_MESSAGE()
{
print STDERR <<"EOM"

$0: Obtain information about libraries, commands and interfaces provided by the given component.

Usage: $0 [options] component_name

component_name                Name of the component to process as it should be
                              given to the package manager

-c|--comp-dbname              Name of the component that should be placed into the database
                              (by default <component_name> is used)

-h|--help                     Display this help message

-f|--file filename            Filename where the data for upload will be situated
                              By default, "<component_name>_upload_data" is used

-l|--filelist filename        File containing list of file names that are installed by the component.
                              (Directory names are also allowed; all files from given directories will be processed)

-v|--version version          Component version. By default, the script will try to obtain version using
                              package manager, but that version might be not that one you'd like to see if
                              the component is provided by several packages, since each package may have
                              its own version

--java                        Collect information about java components
                              (this data is not collected by default)

--distromode                  Work into the 'distribution' mode. In this mode, the script
                              Doesn't analyze dependencies of scripts and modules written in interpreted
                              languages and dependencies of executable files. If this option is specified,
                              only names of modules and commands provided by the component will be collected

-a|--arch                     Architecture name where the component is processed
                              (if it is not set, the output of 'uname' will be used)

-m|--packagemanager pkg_name  Package manager used in the distribution. Supported values
                              are 'rpm', 'dpkg', 'emerge' and 'pacman'. If none is specified, the script
                              will try to guess it.

-q|--quiet                    Be really quiet - do not produce any messages except critical errors

--verbose                     Be verbose

EOM
      ; exit(1);
}
