#!/usr/bin/perl -w

# Linux Application Checker
# Update script (update.pl)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use strict;
use FindBin;

sub BEGIN { push @INC, $FindBin::Bin; }

use Update;

my $verbose = 0;

my $dest_dir;

my $last = 0;
my $check = 0;
my $update = 0;

our @all_archs = qw/x86 x86-64 ia64 ppc32 ppc64 s390 s390x/;
#-----------------------------------------------------------------------

sub fail {
	die "ERROR: ".(join "", @_);
}

sub dbg_out {
	return if !$verbose;
	
	print "* ", @_, "\n";
}
$Update::_dbg_out = \&dbg_out;

$Update::_complain = sub { print STDERR "ERROR: ", @_, "\n" };

#-----------------------------------------------------------------------

# Parse Command Line Args
my @args = @ARGV;
while ( @args ) {
	my $arg = shift @args;
	
	if ( $arg eq '--mk-vers' ) {
		my $loc = shift @args;
		
		print_dir_info($loc) or exit 1;
		exit 0;
	}
	elsif ( $arg eq '--last' ) {
		$last = 1;
	}
	elsif ( $arg eq '--check' ) {
		$check = 1;
	}
	elsif ( $arg eq '--update' ) {
		$update = 1;
	}
	elsif ( $arg eq '--upgrade' ) {
		$check = 1;
		$update = 1;
	}
	elsif ( $arg eq '-v' ) {
		$verbose = 1;
	}
	elsif ( $arg eq '-h' || $arg eq '--help' ) {
		print_help();
		exit;
	}
	elsif ( $arg eq '--dl' ) {
		$Update::dl = shift @args;
		( $Update::dl eq 'wget' || $Update::dl eq 'curl' )
			or fail 'Expected "wget" or "curl" after --dl.';
	}
	elsif ( $arg eq '--dl-opt' ) {
		$Update::dl_opt = shift @args;
	}
	elsif ( $arg eq '--proxy' ) {
		$Update::dl_proxy = shift @args;
	}
	elsif ( $arg eq '--temp-dir' ) {
		$Update::temp_dir = shift @args;
	}
	elsif ( $arg eq '--arch' ) {
		$Update::arch = shift @args;
		( grep {$_ eq $Update::arch} @all_archs )
			or fail "Wrong architecture. Expected one of: @all_archs";
	}
	elsif ( !$dest_dir && $arg !~ /^-/ ) {
		$dest_dir = $arg;
	}
	else {
		print STDERR "Unknown parameter: '$arg'\n\n";
		print_help();
		exit 1;
	}
}

if ( !@ARGV ) {
	print_help();
	exit 0;
}

if ( !$dest_dir ) {
	print STDERR "<dest_dir> expected.\n";
	print_help();
	exit 1;
}

if ( $last ) {
	my $res = db_get_version($dest_dir);
	if ( $res ) {
		print "$res\n";
	} else {
		print "NEVER\n";
	}
}

if ( $check ) {
	my ($res, $size) = db_check_update($dest_dir);
	if ( $res ) {
		print "UPDATE: $res; Download: $size bytes.\n";
	} 
	elsif ( defined $res && $res eq 0 ) {
		print "Everything is already up to date.\n";
	}
}

if ( $update ) {
	my $res = db_update( $dest_dir, sub{print "TOTAL: $_[0]%; $_[1]: $_[2]%\n"} );
	if ( $res ) {
		dbg_out "Done.";
	}
	elsif ( defined $res ) {
		print "Everything is already up to date.\n";
	}
	else {
		print "Error.\n";
	}
}

exit 0;

sub print_help {
	print <<HEREDOC;
Update script for the Linux Application Checker.
DB Version: ${Update::db_ver};
Source: ${Update::source};

Usage: $0 [options...] <dest_dir> [options...]
  --check               Check for updates.
  -h, --help            Show this help.
  --dl <util>           Utility used for downloading files ('wget' or 'curl').
  --dl-opt <options>    Custom options for network downloader. May be used to set up proxy.
  --last                Show last check date.
  --mk-vers <dir>       Make VERSIONS file for files in <dir>.
  --proxy <proxy>       Proxy setup ( [user[:passwd]@]host[:port] )
  --temp-dir <dir>      Use <dir> as a temporary directory for downloading and unpacking.
  --update              Update files in <dir>.
  --upgrade             Check for updates and immediately update if necessary.
  -v                    Verbose output.
HEREDOC
}


sub read_dir_info {
	my ($dir) = @_;
	
	my $data = []; # [ [$filename, $t_ver ,$size],.. ]

	dbg_out "Reading dir '$dir'";
	opendir DIR, $dir  or fail "Failed to open dir '$dir'";
	my @dir_list = readdir(DIR) or fail "Failed to read dir '$dir'";
	closedir DIR;

	my $re_dbfile = join '|', @Update::db_files;
	
	foreach my $filename ( sort @dir_list ) {
		# skip hidden files, '.' and '..'
		next if $filename !~ /^($re_dbfile)_($Update::arch)\.gz$/;
		
		# '-f' checks it is a file
		if (-f "$dir/$filename") {
			my ($size, $mtime) = (stat "$dir/$filename")[7,9];
			
			my $file_ver = Update::timestamp( gmtime($mtime) );
			
			dbg_out "Add file: '$filename' $file_ver $size";
			push @$data, [$filename, $file_ver ,$size];
		}
	}
	
	return $data;
}

sub print_dir_info {
	my ($dir) = @_;
	
	my $data = read_dir_info($dir);
	
	my $versions = { VERSION => $Update::db_ver };
	
	foreach my $arr ( @$data ) {
		my ($filename, $t_ver ,$size) = @$arr;
		
		$versions->{FILES}->{$filename} = { VER => $t_ver, SIZE => $size };
	}
	
	my $ver_file = "VERSIONS_".$Update::arch;
	Update::write_versions($versions, "$dir/$ver_file") or return undef;
}
